
#ifndef PLS_ALIGNED_STACK_H
#define PLS_ALIGNED_STACK_H

#include <cstdint>
#include <cstdlib>

namespace pls {
    namespace internal {
        namespace base {
            class aligned_stack {
                // Keep bounds of our memory block
                char* memory_start_;
                char* memory_end_;

                // Current head will always be aligned to cache lines
                char* head_;

                static std::uintptr_t next_alignment(std::uintptr_t size);
                static char* next_alignment(char* pointer);
            public:
                explicit aligned_stack(char* memory_region, const std::size_t size):
                    memory_start_{memory_region},
                    memory_end_{memory_region + size},
                    head_{next_alignment(memory_start_)} {}

                template<typename T>
                T* push(T object) {
                    T* result = reinterpret_cast<T*>(head_);

                    // Move head to next aligned position after new object
                    head_ = next_alignment(head_ + sizeof(T));
                    if (head_ >= memory_end_) {
                        exit(1); // TODO: Exception Handling
                    }

                    *result = object;
                    return result;
                }

                template<typename T>
                T pop() {
                    head_ = head_ - next_alignment(sizeof(T));

                    return *reinterpret_cast<T*>(head_);
                }
            };
        }
    }
}

#endif //PLS_ALIGNED_STACK_H
